from typing import List, Mapping, Union
from ..action.base_action import BaseAction
from ..task.base_task import BaseTask, Group as TaskGroup
import click

CliSubcommand = Union[click.Group, click.Command]


class Runner(BaseAction):
    '''
    Runner class.
    Any tasks registered to the runner will be accessible from the terminal
    '''

    def __init__(self, env_prefix: str = ''):
        BaseAction.__init__(self)
        self.env_prefix = env_prefix
        self._registered_groups: Mapping[str, click.Group] = {}
        self._top_levels: List[CliSubcommand] = []
        self._subcommands: Mapping[str, List[click.Group]] = {}

    def register(self, task: BaseTask):
        BaseAction.register(self, task)
        task.set_has_cli_interface()

    def serve(self, cli: click.Group) -> click.Group:
        for task in self._tasks:
            subcommand = self._create_cli_subcommand(task)
            if subcommand not in self._top_levels:
                self._top_levels.append(subcommand)
                cli.add_command(subcommand)
        return cli

    def _create_cli_subcommand(self, task: BaseTask) -> click.Group:
        subcommand: CliSubcommand = self._create_cli_command(task)
        task_group = task.group
        while task_group is not None:
            group = self._register_sub_command(task_group, subcommand)
            if task_group.parent is None:
                return group
            subcommand = group
            task_group = task_group.parent
        return subcommand

    def _register_sub_command(
        self, task_group: TaskGroup, subcommand: CliSubcommand
    ) -> click.Group:
        task_group_id = task_group.get_id()
        group = self._get_cli_group(task_group)
        if task_group_id not in self._subcommands:
            self._subcommands[task_group_id] = []
        if subcommand not in self._subcommands[task_group_id]:
            group.add_command(subcommand)
            self._subcommands[task_group_id].append(subcommand)
        return group

    def _get_cli_group(self, task_group: TaskGroup) -> click.Group:
        task_group_id = task_group.get_id()
        if task_group_id in self._registered_groups:
            return self._registered_groups[task_group_id]
        group_cmd_name = task_group.get_cmd_name()
        group_description = task_group.description
        group = click.Group(name=group_cmd_name, help=group_description)
        self._registered_groups[task_group_id] = group
        return group

    def _create_cli_command(self, task: BaseTask) -> click.Command:
        task_inputs = task.get_all_inputs()
        task_cmd_name = task.get_cmd_name()
        task_description = task.get_description()
        task_main_loop = task.create_main_loop(
            env_prefix=self.env_prefix, raise_error=True
         )
        command = click.Command(
            callback=task_main_loop, name=task_cmd_name, help=task_description
        )
        # by default, add an argument named _args
        command.params.append(click.Argument(['_args'], nargs=-1))
        # add task inputs,
        # if there are inputs with the same name, choose the first.
        registered_input: Mapping[str, bool] = {}
        for task_input in task_inputs:
            if task_input.name in registered_input:
                continue
            registered_input[task_input.name] = True
            param_decl = task_input.get_param_decl()
            options = task_input.get_options()
            command.params.append(click.Option(param_decl, **options))
        return command



from module.auth.core.authorizer.authorizer import Authorizer
from core.rpc import Caller


class RPCAuthorizer(Authorizer):

    def __init__(
        self,
        rpc_caller: Caller,
        is_admin_rpc_name: str,
        is_guest_rpc_name: str,
        is_having_permission_rpc_name: str,
    ):
        self.rpc_caller = rpc_caller
        self.is_admin_rpc_name = is_admin_rpc_name
        self.is_guest_rpc_name = is_guest_rpc_name
        self.is_having_permission_rpc_name = is_having_permission_rpc_name

    async def is_admin(self, user_id: str) -> bool:
        return await self.rpc_caller.call(
            self.is_admin_rpc_name, user_id
        )

    async def is_guest(self, user_id: str) -> bool:
        return await self.rpc_caller.call(
            self.is_guest_rpc_name, user_id
        )

    async def is_having_permission(
        self, user_id: str, permission_name
    ) -> bool:
        return await self.rpc_caller.call(
            self.is_having_permission_rpc_name, user_id, permission_name
        )



from typing import Annotated
from fastapi import FastAPI, Depends
from fastapi.security import OAuth2PasswordRequestForm
from logging import Logger
from core.messagebus import Publisher
from core.rpc import Caller
from core.error import HTTPAPIException
from module.auth.core import Authorizer
from module.auth.schema.user import (
    User, UserData, UserResult, UserLogin
)
from module.auth.schema.token import TokenData, TokenResponse
from module.auth.component import token_scheme


def register_login_api(
    logger: Logger,
    app: FastAPI,
    authorizer: Authorizer,
    rpc_caller: Caller,
    publisher: Publisher
):
    logger.info('🥪 Register Login API for "auth.user"')

    @app.post('/api/v1/login-oauth', response_model=TokenResponse)
    async def login_oauth(
        form_data: Annotated[OAuth2PasswordRequestForm, Depends()]
    ) -> TokenResponse:
        data = UserLogin(
            identity=form_data.username,
            password=form_data.password
        )
        return await _login(data=data)

    @app.post('/api/v1/login', response_model=TokenResponse)
    async def login(data: UserLogin) -> TokenResponse:
        return await _login(data=data)

    async def _login(data: UserLogin) -> TokenResponse:
        try:
            token = await rpc_caller.call(
                'auth_login', login_data=data.dict()
            )
            return TokenResponse(access_token=token, token_type='bearer')
        except Exception as e:
            raise HTTPAPIException(error=e)


def register_api(
    logger: Logger,
    app: FastAPI,
    authorizer: Authorizer,
    rpc_caller: Caller,
    publisher: Publisher
):
    logger.info('🥪 Register API for "auth.user"')

    @app.get(
        '/api/v1/auth/users', response_model=UserResult
    )
    async def get_users(
        keyword: str = '', limit: int = 100, offset: int = 0,
        user_token_data: TokenData = Depends(token_scheme)
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, 'auth:user:get'
        ):
            raise HTTPAPIException(403, 'Unauthorized')
        try:
            result_dict = await rpc_caller.call(
                'auth_get_user',
                keyword=keyword,
                criterion={},
                limit=limit,
                offset=offset,
                user_token_data=user_token_data.dict()
            )
            return UserResult(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.get(
        '/api/v1/auth/users/{id}', response_model=User
    )
    async def get_user_by_id(
        id: str, user_token_data: TokenData = Depends(token_scheme)
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, 'auth:user:get_by_id'
        ):
            raise HTTPAPIException(403, 'Unauthorized')
        try:
            result_dict = await rpc_caller.call(
                'auth_get_user_by_id',
                id=id, user_token_data=user_token_data.dict()
            )
            return User(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.post(
        '/api/v1/auth/users', response_model=User
    )
    async def insert_user(
        data: UserData, user_token_data: TokenData = Depends(token_scheme)
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, 'auth:user:insert'
        ):
            raise HTTPAPIException(403, 'Unauthorized')
        try:
            result_dict = await rpc_caller.call(
                'auth_insert_user',
                data=data.dict(), user_token_data=user_token_data.dict()
            )
            return User(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.put(
        '/api/v1/auth/users/{id}', response_model=User
    )
    async def update_user(
        id: str, data: UserData,
        user_token_data: TokenData = Depends(token_scheme)
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, 'auth:user:update'
        ):
            raise HTTPAPIException(403, 'Unauthorized')
        try:
            result_dict = await rpc_caller.call(
                'auth_update_user',
                id=id, data=data.dict(), user_token_data=user_token_data.dict()
            )
            return User(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.delete(
        '/api/v1/auth/users/{id}', response_model=User
    )
    async def delete_user(
        id: str, user_token_data: TokenData = Depends(token_scheme)
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, 'auth:user:delete'
        ):
            raise HTTPAPIException(403, 'Unauthorized')
        try:
            result_dict = await rpc_caller.call(
                'auth_delete_user',
                id=id, user_token_data=user_token_data.dict()
            )
            return User(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)



from typing import Optional
from core.model import RepoModel
from module.auth.schema.user import (
    User, UserData, UserResult, UserLogin
)
from module.auth.entity.user.repo import UserRepo
from module.auth.schema.token import TokenData
from module.auth.core import TokenUtil
from module.auth.entity.permission.model import PermissionModel


class UserModel(
    RepoModel[User, UserData, UserResult]
):
    schema_result_cls = UserResult

    def __init__(
        self,
        repo: UserRepo,
        permission_model: PermissionModel,
        token_util: TokenUtil,
        expire_seconds: int,
        guest_user: User,
        admin_user: Optional[User] = None,
        admin_user_password: str = ''
    ):
        self.repo = repo
        self.permission_model = permission_model
        self.token_util = token_util
        self.expire_seconds = expire_seconds
        self.guest_user = guest_user
        self.admin_user = admin_user
        self.admin_user_pasword = admin_user_password

    async def insert(self, data: UserData) -> User:
        if (
            self.admin_user is not None and
            data.username == self.admin_user.username
        ):
            raise ValueError(
                'Invalid username: {data.username} is used by admin'
            )
        if data.username == self.guest_user.username:
            raise ValueError(
                'Invalid username: {data.username} is used by guest'
            )
        return await super().insert(data)

    async def update(self, id: str, data: UserData) -> User:
        if self.admin_user is not None:
            if id == self.admin_user.id:
                raise ValueError(
                    'Forbidden: editting admin user is not permitted'
                )
            if data.username == self.admin_user.username:
                raise ValueError(
                    'Invalid username: {data.username} is used by admin'
                )
        if id == self.guest_user.id:
            raise ValueError(
                'Forbidden: editting guest user is not permitted'
            )
        if data.username == self.guest_user.username:
            raise ValueError(
                'Invalid username: {data.username} is used by guest'
            )
        return await super().update(id, data)

    async def delete(self, id: str) -> User:
        if (
            self.admin_user is not None and
            id == self.admin_user.id
        ):
            raise ValueError(
                'Forbidden: deleting admin user is not permitted'
            )
        if id == self.guest_user.id:
            raise ValueError(
                'Forbidden: deleting guest user is not permitted'
            )
        return await super().delete(id)

    async def is_having_permission(
        self, id: str, permission_name: str
    ) -> bool:
        if await self.is_admin(id):
            return True
        if self.is_guest(id):
            return self._is_user_having_permission(
                self.guest_user, permission_name
            )
        user = await self.get_by_id(id)
        return self._is_user_having_permission(user, permission_name)

    async def is_guest(self, id: str) -> bool:
        return id == self.guest_user.id

    async def is_admin(self, id: str) -> bool:
        if self.admin_user is None:
            return False
        return id == self.admin_user.id

    async def login(self, user_login: UserLogin) -> str:
        if user_login.identity == '':
            raise ValueError('Invalid identity: Identity is not provided')
        if (
            self.admin_user is not None
            and user_login.password == self.admin_user_pasword
            and (
                user_login.identity == self.admin_user.username or
                user_login.identity == self.admin_user.email or
                user_login.identity == self.admin_user.phone
            )
        ):
            return self._get_token(self.admin_user)
        user = await self._get_user_by_user_login(user_login)
        return self._get_token(user)

    def _is_user_having_permission(
        self, user: User, permission_name: str
    ) -> bool:
        user_permission_names = [
            permission.name for permission in user.permissions
        ]
        if permission_name in user_permission_names:
            return True
        for group in user.groups:
            group_permission_names = [
                permission.name for permission in group.permissions
            ]
            if permission_name in group_permission_names:
                return True
        return False

    def _get_token(self, user: User) -> str:
        token_data = TokenData(
            user_id=user.id,
            username=user.username,
            expire_seconds=self.expire_seconds
        )
        return self.token_util.encode(token_data)

    async def _get_user_by_user_login(self, user_login: UserLogin) -> User:
        return await self.repo.get_by_user_login(user_login)
